@@ -2,18 +2,21 @@ package btf
 
 import (
 	"errors"
-	"math/rand"
+	"fmt"
 	"os"
+	"slices"
 	"strings"
 	"testing"
 
-	"github.com/cilium/ebpf/internal/testutils"
 	"github.com/google/go-cmp/cmp"
 
-	qt "github.com/frankban/quicktest"
+	"github.com/cilium/ebpf/internal"
+	"github.com/cilium/ebpf/internal/testutils"
+
+	"github.com/go-quicktest/qt"
 )
 
-func TestCOREAreTypesCompatible(t *testing.T) {
+func TestCheckTypeCompatibility(t *testing.T) {
 	tests := []struct {
 		a, b       Type
 		compatible bool
@@ -45,38 +48,40 @@ func TestCOREAreTypesCompatible(t *testing.T) {
 			&FuncProto{Return: &Void{}, Params: []FuncParam{{Type: &Void{}}}},
 			false,
 		},
+		{&FuncProto{Return: &Typedef{Type: &Int{}}}, &FuncProto{Return: &Int{}}, true},
+		{&FuncProto{Return: &Typedef{Type: &Int{}}}, &FuncProto{Return: &Void{}}, false},
 	}
 
 	for _, test := range tests {
-		err := coreAreTypesCompatible(test.a, test.b)
+		err := CheckTypeCompatibility(test.a, test.b)
 		if test.compatible {
 			if err != nil {
 				t.Errorf("Expected types to be compatible: %s\na = %#v\nb = %#v", err, test.a, test.b)
 				continue
 			}
 		} else {
-			if !errors.Is(err, errImpossibleRelocation) {
+			if !errors.Is(err, errIncompatibleTypes) {
 				t.Errorf("Expected types to be incompatible: %s\na = %#v\nb = %#v", err, test.a, test.b)
 				continue
 			}
 		}
 
-		err = coreAreTypesCompatible(test.b, test.a)
+		err = CheckTypeCompatibility(test.b, test.a)
 		if test.compatible {
 			if err != nil {
 				t.Errorf("Expected reversed types to be compatible: %s\na = %#v\nb = %#v", err, test.a, test.b)
 			}
 		} else {
-			if !errors.Is(err, errImpossibleRelocation) {
+			if !errors.Is(err, errIncompatibleTypes) {
 				t.Errorf("Expected reversed types to be incompatible: %s\na = %#v\nb = %#v", err, test.a, test.b)
 			}
 		}
 	}
 
 	for _, invalid := range []Type{&Var{}, &Datasec{}} {
-		err := coreAreTypesCompatible(invalid, invalid)
-		if errors.Is(err, errImpossibleRelocation) {
-			t.Errorf("Expected an error for %T, not errImpossibleRelocation", invalid)
+		err := CheckTypeCompatibility(invalid, invalid)
+		if errors.Is(err, errIncompatibleTypes) {
+			t.Errorf("Expected an error for %T, not errIncompatibleTypes", invalid)
 		} else if err == nil {
 			t.Errorf("Expected an error for %T", invalid)
 		}
@@ -209,7 +214,7 @@ func TestCOREFindEnumValue(t *testing.T) {
 		name                    string
 		local, target           Type
 		acc                     coreAccessor
-		localValue, targetValue int32
+		localValue, targetValue uint64
 	}{
 		{"a to b", a, b, coreAccessor{0}, 23, 0},
 		{"b to a", b, a, coreAccessor{1}, 123, 42},
@@ -218,9 +223,9 @@ func TestCOREFindEnumValue(t *testing.T) {
 	for _, test := range valid {
 		t.Run(test.name, func(t *testing.T) {
 			local, target, err := coreFindEnumValue(test.local, test.acc, test.target)
-			qt.Assert(t, err, qt.IsNil)
-			qt.Check(t, local.Value, qt.Equals, test.localValue)
-			qt.Check(t, target.Value, qt.Equals, test.targetValue)
+			qt.Assert(t, qt.IsNil(err))
+			qt.Check(t, qt.Equals(local.Value, test.localValue))
+			qt.Check(t, qt.Equals(target.Value, test.targetValue))
 		})
 	}
 }
@@ -302,6 +307,13 @@ func TestCOREFindField(t *testing.T) {
 			coreAccessor{0, 0},
 			errImpossibleRelocation,
 		},
+		{
+			"unsized type",
+			bStruct, &Func{},
+			// non-zero accessor to force calculating the offset.
+			coreAccessor{1},
+			errImpossibleRelocation,
+		},
 	}
 
 	for _, test := range invalid {
@@ -478,7 +490,7 @@ func TestCOREFindField(t *testing.T) {
 	for _, test := range valid {
 		t.Run(test.name, func(t *testing.T) {
 			localField, targetField, err := coreFindField(test.local, test.acc, test.target)
-			qt.Assert(t, err, qt.IsNil)
+			qt.Assert(t, qt.IsNil(err))
 			checkCOREField(t, "local", localField, test.localField)
 			checkCOREField(t, "target", targetField, test.targetField)
 		})
@@ -539,14 +551,14 @@ func TestCORERelocation(t *testing.T) {
 		}
 
 		for section := range extInfos.funcInfos {
-			name := strings.TrimPrefix(section, "socket_filter/")
+			name := strings.TrimPrefix(section, "socket/")
 			t.Run(name, func(t *testing.T) {
 				var relos []*CORERelocation
-				for _, reloInfo := range extInfos.relocationInfos[section] {
+				for _, reloInfo := range extInfos.relocationInfos[section].infos {
 					relos = append(relos, reloInfo.relo)
 				}
 
-				fixups, err := CORERelocate(spec, spec, relos)
+				fixups, err := CORERelocate(relos, []*Spec{spec}, spec.byteOrder, spec.TypeID)
 				if want := errs[name]; want != nil {
 					if !errors.Is(err, want) {
 						t.Fatal("Expected", want, "got", err)
@@ -570,50 +582,354 @@ func TestCORERelocation(t *testing.T) {
 	})
 }
 
-func TestCORECopyWithoutQualifiers(t *testing.T) {
-	qualifiers := []struct {
-		name string
-		fn   func(Type) Type
-	}{
-		{"const", func(t Type) Type { return &Const{Type: t} }},
-		{"volatile", func(t Type) Type { return &Volatile{Type: t} }},
-		{"restrict", func(t Type) Type { return &Restrict{Type: t} }},
-		{"typedef", func(t Type) Type { return &Typedef{Type: t} }},
-	}
-
-	for _, test := range qualifiers {
-		t.Run(test.name+" cycle", func(t *testing.T) {
-			root := &Volatile{}
-			root.Type = test.fn(root)
-
-			cycle, ok := Copy(root, UnderlyingType).(*cycle)
-			qt.Assert(t, ok, qt.IsTrue)
-			qt.Assert(t, cycle.root, qt.Equals, root)
+func TestCOREReloFieldSigned(t *testing.T) {
+	for _, typ := range []Type{&Int{}, &Enum{}} {
+		t.Run(fmt.Sprintf("%T with invalid target", typ), func(t *testing.T) {
+			relo := &CORERelocation{
+				typ, coreAccessor{0}, reloFieldSigned, 0,
+			}
+			fixup, err := coreCalculateFixup(relo, &Void{}, internal.NativeEndian, dummyTypeID)
+			qt.Assert(t, qt.IsTrue(fixup.poison))
+			qt.Assert(t, qt.IsNil(err))
 		})
 	}
 
-	for _, a := range qualifiers {
-		for _, b := range qualifiers {
-			t.Run(a.name+" "+b.name, func(t *testing.T) {
-				v := a.fn(&Pointer{Target: b.fn(&Int{Name: "z"})})
-				want := &Pointer{Target: &Int{Name: "z"}}
+	t.Run("type without signedness", func(t *testing.T) {
+		relo := &CORERelocation{
+			&Array{}, coreAccessor{0}, reloFieldSigned, 0,
+		}
+		_, err := coreCalculateFixup(relo, &Array{}, internal.NativeEndian, dummyTypeID)
+		qt.Assert(t, qt.ErrorIs(err, errNoSignedness))
+	})
+}
 
-				got := Copy(v, UnderlyingType)
-				qt.Assert(t, got, qt.DeepEquals, want)
-			})
+func TestCOREReloFieldShiftU64(t *testing.T) {
+	typ := &Struct{
+		Members: []Member{
+			{Name: "A", Type: &Fwd{}},
+		},
+	}
+
+	for _, relo := range []*CORERelocation{
+		{typ, coreAccessor{0, 0}, reloFieldRShiftU64, 1},
+		{typ, coreAccessor{0, 0}, reloFieldLShiftU64, 1},
+	} {
+		t.Run(relo.kind.String(), func(t *testing.T) {
+			_, err := coreCalculateFixup(relo, typ, internal.NativeEndian, dummyTypeID)
+			qt.Assert(t, qt.ErrorIs(err, errUnsizedType))
+		})
+	}
+}
+
+func TestCORERelosKmodTypeID(t *testing.T) {
+	a := &Int{Name: "a"}
+	b := &Int{Name: "b"}
+
+	relos := []*CORERelocation{
+		{&Int{}, coreAccessor{0}, reloTypeIDTarget, 0},
+	}
+
+	typeID := func(t Type) (TypeID, error) {
+		if t == a {
+			return 42, nil
+		}
+		return 0, ErrNotFound
+	}
+
+	fixups, err := coreCalculateFixups(
+		relos,
+		[]Type{a, b},
+		internal.NativeEndian,
+		typeID,
+	)
+	qt.Assert(t, qt.IsNil(err))
+	qt.Assert(t, qt.IsFalse(fixups[0].poison))
+	qt.Assert(t, qt.Equals(fixups[0].target, 42))
+
+	fixups, err = coreCalculateFixups(
+		relos,
+		[]Type{b},
+		internal.NativeEndian,
+		typeID,
+	)
+	qt.Assert(t, qt.IsNil(err))
+	qt.Assert(t, qt.IsTrue(fixups[0].poison))
+}
+
+func BenchmarkCORESkBuff(b *testing.B) {
+	spec := vmlinuxTestdataSpec(b)
+
+	var skb *Struct
+	err := spec.TypeByName("sk_buff", &skb)
+	qt.Assert(b, qt.IsNil(err))
+
+	skbID, err := spec.TypeID(skb)
+	qt.Assert(b, qt.IsNil(err))
+
+	lenIndex := slices.IndexFunc(skb.Members, func(m Member) bool {
+		return m.Name == "len"
+	})
+	qt.Assert(b, qt.Not(qt.Equals(lenIndex, -1)))
+
+	var pktHashTypes *Enum
+	err = spec.TypeByName("pkt_hash_types", &pktHashTypes)
+	qt.Assert(b, qt.IsNil(err))
+
+	pktHashTypesID, err := spec.TypeID(pktHashTypes)
+	qt.Assert(b, qt.IsNil(err))
+
+	for _, relo := range []*CORERelocation{
+		{skb, coreAccessor{0, lenIndex}, reloFieldByteOffset, skbID},
+		{skb, coreAccessor{0, lenIndex}, reloFieldByteSize, skbID},
+		{skb, coreAccessor{0, lenIndex}, reloFieldExists, skbID},
+		{skb, coreAccessor{0, lenIndex}, reloFieldSigned, skbID},
+		{skb, coreAccessor{0, lenIndex}, reloFieldLShiftU64, skbID},
+		{skb, coreAccessor{0, lenIndex}, reloFieldRShiftU64, skbID},
+		{skb, coreAccessor{0}, reloTypeIDLocal, skbID},
+		{skb, coreAccessor{0}, reloTypeIDTarget, skbID},
+		{skb, coreAccessor{0}, reloTypeExists, skbID},
+		{skb, coreAccessor{0}, reloTypeSize, skbID},
+		{pktHashTypes, coreAccessor{0}, reloEnumvalExists, pktHashTypesID},
+		{pktHashTypes, coreAccessor{0}, reloEnumvalValue, pktHashTypesID},
+	} {
+		b.Run(relo.kind.String(), func(b *testing.B) {
+			b.ReportAllocs()
+
+			for b.Loop() {
+				_, err = CORERelocate([]*CORERelocation{relo}, []*Spec{spec}, spec.byteOrder, spec.TypeID)
+				if err != nil {
+					b.Fatal(err)
+				}
+			}
+		})
+	}
+}
+
+func TestCORETypesMatch(t *testing.T) {
+	tests := []struct {
+		a, b       Type
+		match      bool
+		reversible bool
+	}{
+		{&Void{}, &Void{}, true, true},
+		{&Int{Size: 32}, &Int{Size: 32}, true, true},
+		{&Int{Size: 64}, &Int{Size: 32}, false, true},
+		{&Int{Size: 32}, &Int{Size: 32, Encoding: Signed}, false, true},
+		{&Fwd{Name: "a"}, &Fwd{Name: "a"}, true, true},
+		{&Fwd{Name: "a"}, &Fwd{Name: "b___new"}, false, true},
+		{&Fwd{Name: "a"}, &Fwd{Name: "a___new"}, true, true},
+		{&Fwd{Name: "a"}, &Struct{Name: "a___new"}, false, true},
+		{&Fwd{Name: "a"}, &Union{Name: "a___new"}, false, true},
+		{&Fwd{Name: "a", Kind: FwdStruct}, &Fwd{Name: "a___new", Kind: FwdUnion}, false, true},
+		{&Pointer{&Fwd{Name: "a", Kind: FwdStruct}}, &Pointer{&Struct{Name: "a___new"}}, true, true},
+		{&Pointer{&Fwd{Name: "a", Kind: FwdUnion}}, &Pointer{&Union{Name: "a___new"}}, true, true},
+		{&Pointer{&Fwd{Name: "a", Kind: FwdStruct}}, &Pointer{&Union{Name: "a___new"}}, false, true},
+		{&Struct{Name: "a___new"}, &Union{Name: "a___new"}, false, true},
+		{&Pointer{&Struct{Name: "a"}}, &Pointer{&Union{Name: "a___new"}}, false, true},
+		{
+			&Struct{Name: "a", Members: []Member{
+				{Name: "foo", Type: &Int{}},
+			}},
+			&Struct{Name: "a___new", Members: []Member{
+				{Name: "foo", Type: &Int{}},
+			}},
+			true,
+			true,
+		},
+		{
+			&Struct{Name: "a", Members: []Member{
+				{Name: "foo", Type: &Int{}},
+			}},
+			&Struct{Name: "a___new", Members: []Member{
+				{Name: "foo", Type: &Int{}},
+				{Name: "bar", Type: &Int{}},
+			}},
+			true,
+			false,
+		},
+		{
+			&Struct{Name: "a", Members: []Member{
+				{Name: "foo", Type: &Int{}},
+				{Name: "bar", Type: &Int{}},
+			}},
+			&Struct{Name: "a___new", Members: []Member{
+				{Name: "foo", Type: &Int{}},
+			}},
+			false,
+			false,
+		},
+		{
+			&Struct{Name: "a", Members: []Member{
+				{Name: "bar", Type: &Int{}},
+			}},
+			&Struct{Name: "a___new", Members: []Member{
+				{Name: "foo", Type: &Int{}},
+			}},
+			false,
+			false,
+		},
+		{
+			&Struct{Name: "a", Members: []Member{
+				{Name: "foo", Type: &Int{Encoding: Signed}},
+			}},
+			&Struct{Name: "a___new", Members: []Member{
+				{Name: "foo", Type: &Int{}},
+			}},
+			false,
+			false,
+		},
+		{
+			&Enum{Name: "a", Values: []EnumValue{
+				{"foo", 1},
+			}},
+			&Enum{Name: "a___new", Values: []EnumValue{
+				{"foo", 1},
+			}},
+			true,
+			true,
+		},
+		{
+			&Enum{Name: "a", Values: []EnumValue{
+				{"foo", 1},
+			}},
+			&Enum{Name: "a___new", Values: []EnumValue{
+				{"foo", 1},
+				{"bar", 2},
+			}},
+			true,
+			false,
+		},
+		{
+			&Enum{Name: "a", Values: []EnumValue{
+				{"foo", 1},
+				{"bar", 2},
+			}},
+			&Enum{Name: "a___new", Values: []EnumValue{
+				{"foo", 1},
+			}},
+			false,
+			false,
+		},
+		{
+			&Enum{Name: "a", Values: []EnumValue{
+				{"foo", 1},
+			}},
+			&Enum{Name: "a___new", Values: []EnumValue{
+				{"bar", 1},
+			}},
+			false,
+			false,
+		},
+		{
+			&Enum{Name: "a", Values: []EnumValue{
+				{"foo", 1},
+			}, Size: 1},
+			&Enum{Name: "a___new", Values: []EnumValue{
+				{"foo", 1},
+			}, Size: 2},
+			false,
+			false,
+		},
+		{
+			&Array{Type: &Int{}, Nelems: 2},
+			&Array{Type: &Int{}, Nelems: 2},
+			true,
+			true,
+		},
+		{
+			&Array{Type: &Int{}, Nelems: 3},
+			&Array{Type: &Int{}, Nelems: 2},
+			false,
+			true,
+		},
+		{
+			&Array{Type: &Void{}, Nelems: 2},
+			&Array{Type: &Int{}, Nelems: 2},
+			false,
+			true,
+		},
+		{
+			&FuncProto{Return: &Int{}, Params: []FuncParam{
+				{Name: "foo", Type: &Int{}},
+			}},
+			&FuncProto{Return: &Int{}, Params: []FuncParam{
+				{Name: "bar", Type: &Int{}},
+			}},
+			true,
+			true,
+		},
+		{
+			&FuncProto{Return: &Int{}, Params: []FuncParam{
+				{Name: "foo", Type: &Int{}},
+			}},
+			&FuncProto{Return: &Int{}, Params: []FuncParam{
+				{Name: "bar", Type: &Int{}},
+				{Name: "baz", Type: &Int{}},
+			}},
+			false,
+			true,
+		},
+		{
+			&FuncProto{Return: &Void{}, Params: []FuncParam{
+				{Name: "foo", Type: &Int{}},
+			}},
+			&FuncProto{Return: &Int{}, Params: []FuncParam{
+				{Name: "bar", Type: &Int{}},
+			}},
+			false,
+			true,
+		},
+		{
+			&FuncProto{Return: &Void{}, Params: []FuncParam{
+				{Name: "bar", Type: &Int{Encoding: Signed}},
+			}},
+			&FuncProto{Return: &Int{}, Params: []FuncParam{
+				{Name: "bar", Type: &Int{}},
+			}},
+			false,
+			true,
+		},
+	}
+
+	for _, test := range tests {
+		err := coreTypesMatch(test.a, test.b, nil)
+		if test.match {
+			if err != nil {
+				t.Errorf("Expected types to match: %s\na = %#v\nb = %#v", err, test.a, test.b)
+				continue
+			}
+		} else {
+			if !errors.Is(err, errIncompatibleTypes) {
+				t.Errorf("Expected types to be incompatible: \na = %#v\nb = %#v", test.a, test.b)
+				continue
+			}
+		}
+
+		if test.reversible {
+			err = coreTypesMatch(test.b, test.a, nil)
+			if test.match {
+				if err != nil {
+					t.Errorf("Expected reversed types to match: %s\na = %#v\nb = %#v", err, test.a, test.b)
+				}
+			} else {
+				if !errors.Is(err, errIncompatibleTypes) {
+					t.Errorf("Expected reversed types to be incompatible: %s\na = %#v\nb = %#v", err, test.a, test.b)
+				}
+			}
 		}
 	}
 
-	t.Run("long chain", func(t *testing.T) {
-		root := &Int{Name: "abc"}
-		v := Type(root)
-		for i := 0; i < maxTypeDepth; i++ {
-			q := qualifiers[rand.Intn(len(qualifiers))]
-			v = q.fn(v)
-			t.Log(q.name)
+	for _, invalid := range []Type{&Var{}, &Datasec{}} {
+		err := coreTypesMatch(invalid, invalid, nil)
+		if errors.Is(err, errIncompatibleTypes) {
+			t.Errorf("Expected an error for %T, not errIncompatibleTypes", invalid)
+		} else if err == nil {
+			t.Errorf("Expected an error for %T", invalid)
 		}
+	}
+}
 
-		got := Copy(v, UnderlyingType)
-		qt.Assert(t, got, qt.DeepEquals, root)
-	})
+// dummyTypeID returns 0, nil for any passed type.
+func dummyTypeID(Type) (TypeID, error) {
+	return 0, nil
 }
