@@ -10,8 +10,12 @@ import (
 )
 
 // ContainerTop shows process information from within a container.
-func (cli *Client) ContainerTop(ctx context.Context, containerID string, arguments []string) (container.ContainerTopOKBody, error) {
-	var response container.ContainerTopOKBody
+func (cli *Client) ContainerTop(ctx context.Context, containerID string, arguments []string) (container.TopResponse, error) {
+	containerID, err := trimID("container", containerID)
+	if err != nil {
+		return container.TopResponse{}, err
+	}
+
 	query := url.Values{}
 	if len(arguments) > 0 {
 		query.Set("ps_args", strings.Join(arguments, " "))
@@ -20,9 +24,10 @@ func (cli *Client) ContainerTop(ctx context.Context, containerID string, argumen
 	resp, err := cli.get(ctx, "/containers/"+containerID+"/top", query, nil)
 	defer ensureReaderClosed(resp)
 	if err != nil {
-		return response, err
+		return container.TopResponse{}, err
 	}
 
-	err = json.NewDecoder(resp.body).Decode(&response)
+	var response container.TopResponse
+	err = json.NewDecoder(resp.Body).Decode(&response)
 	return response, err
 }
